from typing import Callable, Dict
from datetime import date

import ibis
from ibis import _

from phenex.phenotypes.phenotype import Phenotype
from phenex.filters.relative_time_range_filter import RelativeTimeRangeFilter
from phenex.filters import DateFilter, ValueFilter
from phenex.tables import is_phenex_code_table, PHENOTYPE_TABLE_COLUMNS, PhenotypeTable
from phenex.aggregators import First, Last

from phenex.util import create_logger

logger = create_logger(__name__)


class UserDefinedPhenotype(Phenotype):
    """
    UserDefinedPhenotype allows users of PhenEx to implement custom functionality within a single phenotype. To use, the user must pass a function that returns an ibis table. This means that the function must
    1. return an ibis table
    2. There are a minimum of one column : PERSON_ID. If no other columns are returned, it is assumed that all person_ids in the PERSON_ID column fulfill the UserDefinedPhenotype
    3. If additional columns are returned, they must be named BOOLEAN, EVENT_DATE, and VALUE. The BOOLEAN column indicates whether the person_id fulfills the UserDefinedPhenotype; patients with BOOLEAN = False will be removed. The EVENT_DATE column contains the date of the event, and the VALUE column contains a numeric value associated with the event. Any other columns are ignored.

    UserDefinedPhenotype is especially useful for two use cases :
    1. Hybrid workflows: If you have performed cohort extraction outside of PhenEx (e.g. in R, SQL) but would like to use PhenEx to calculate baseline characteristics and outcomes, we can set the entry criterion to a UserDefinedPhenotype and read a dataframe of PERSON_IDS and INDEX_DATES. In this way, PhenEx flexibly allows us to use multiple tools in our analysis.
    2. Custom event definitions: If you need to define events based on complex logic that is not easily expressed using the built-in PhenEx functionality, you can use UserDefinedPhenotype to implement this logic in a custom function.

    DATE: custom, as defined by user
    VALUE: custom, as defined by user

    Parameters:
        name: The name of the phenotype.
        function: A function that takes mapped_tables as input and returns a PhenotypeTable.

    Example:
        ```python

        # define a custom function that returns at a minimum the PERSON_ID column
        def custom_function(mapped_tables):
            table = mapped_tables['PERSON']
            table = table.filter(table.AGE > 18)
            table = table.select("PERSON_ID")
            return table

        phenotype = UserDefinedPhenotype(
            name="example_phenotype",
            function=custom_function
        )

        tables = {"PERSON": example_code_table}

        result_table = phenotype.execute(tables)
        display(result_table)
        ```
    """

    def __init__(
        self,
        function: Callable[[Dict[str, "PhenexTable"]], "PhenexTable"],
        **kwargs,
    ):
        super(UserDefinedPhenotype, self).__init__(**kwargs)
        self.function = function

    def _execute(self, tables) -> PhenotypeTable:
        table = self.function(tables)
        import ibis

        ibis.options.interactive = True
        print(table)

        if "BOOLEAN" not in table.columns:
            table = table.mutate(BOOLEAN=True).distinct()
        else:
            table = table.filter(table.BOOLEAN == True)

        if "EVENT_DATE" not in table.columns:
            table = table.mutate(EVENT_DATE=ibis.null(date))
        if "VALUE" not in table.columns:
            table = table.mutate(VALUE=ibis.null().cast("int32"))
        print(table)
        return table



from setuptools import setup, find_packages

__version__ = "v0.6.0"


# Function to parse requirements.txt
def parse_requirements(filename):
    with open(filename, "r") as file:
        return [
            line.strip() for line in file if line.strip() and not line.startswith("#")
        ]


setup(
    name="phenex",
    version=__version__,
    author="Bayer AG",
    author_email="alexander.hartenstein@bayer.com",
    description="PhenEx is a Python package for analysis-ready datasets from real-world data",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Bayer-Group/PhenEx",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.12",
    install_requires=parse_requirements("requirements.txt"),
)



from typing import Optional

# from phenex.phenotypes.phenotype import Phenotype
from phenex.filters.filter import Filter
from phenex.filters.value_filter import ValueFilter
from phenex.tables import EventTable, is_phenex_phenotype_table
from phenex.filters.value import *


class RelativeTimeRangeFilter(Filter):
    """
    This class filters events in an EventTable based on a specified time range relative to an anchor date.  The anchor date can either be provided by an anchor phenotype or by an 'INDEX_DATE' column in the EventTable.

    Parameters:
        min_days: Minimum number of days from the anchor date to filter events.
        max_days: Maximum number of days from the anchor date to filter events.
        anchor_phenotype: A phenotype providing the anchor date for filtering.
        when: when can be "before" or "after"; if "before", days prior to anchor event_date are positive, and days after are negative; using after, days before the anchor event_date are negative and days after the anchor event_date are positive.

    Methods:
        filter: Filters the given EventTable based on the specified time range relative to the anchor date.

    Examples:
        ```
        # filter events to one year before index date, excluding index date
        from phenex.filters.value import LessThan, GreaterThan
        one_year_preindex = RelativeTimeRangeFilter(
            max_days = LessThan(365),
            min_days = GreaterThan(0),
            when = 'before'
            )
        ```

        ```
        # filter events to one year after index date, including index date
        anytime_after_index = RelativeTimeRangeFilter(
            min_days = GreaterThan(0),
            when = 'after'
            )
        ```
    """

    def __init__(
        self,
        min_days: Optional[Value] = GreaterThanOrEqualTo(0),
        max_days: Optional[Value] = None,
        when: Optional[str] = "before",
        anchor_phenotype: Optional["Phenotype"] = None,
    ):
        verify_relative_time_range_filter_input(min_days, max_days, when)

        self.min_days = min_days
        self.max_days = max_days
        self.when = when
        self.anchor_phenotype = anchor_phenotype
        super(RelativeTimeRangeFilter, self).__init__()

    def _filter(self, table: EventTable):
        if self.anchor_phenotype is not None:
            if self.anchor_phenotype.table is None:
                raise ValueError(
                    f"Dependent Phenotype {self.anchor_phenotype.name} must be executed before this node can run!"
                )
            else:
                anchor_table = self.anchor_phenotype.table
                reference_column = anchor_table.EVENT_DATE
                # Note that joins can change column names if the tables have name collisions!
                table = table.join(anchor_table, "PERSON_ID")
        else:
            assert (
                "INDEX_DATE" in table.columns
            ), f"INDEX_DATE column not found in table {table}"
            reference_column = table.INDEX_DATE

        DAYS_FROM_ANCHOR = reference_column.delta(table.EVENT_DATE, "day")
        if self.when == "after":
            DAYS_FROM_ANCHOR = -DAYS_FROM_ANCHOR

        table = table.mutate(DAYS_FROM_ANCHOR=DAYS_FROM_ANCHOR)

        conditions = []

        value_filter = ValueFilter(
            min_value=self.min_days,
            max_value=self.max_days,
            column_name="DAYS_FROM_ANCHOR",
        )

        return value_filter.filter(table)


def verify_relative_time_range_filter_input(min_days, max_days, when):
    if min_days is not None:
        assert min_days.operator in [
            ">",
            ">=",
        ], f"min_days operator must be > or >=, not {min_days.operator}"
    if max_days is not None:
        assert max_days.operator in [
            "<",
            "<=",
        ], f"max_days operator must be > or >=, not {max_days.operator}"
    if max_days is not None and min_days is not None:
        assert (
            min_days.value <= max_days.value
        ), f"min_days must be less than or equal to max_days"
    assert when in [
        "before",
        "after",
    ], f"when must be 'before' or 'after', not {when}"



from typing import Optional, Union
from datetime import date, datetime
from .value_filter import ValueFilter
from .value import Value


class Date(Value):
    """
    The Date class is a specialized Value class for handling date comparisons.

    Attributes:
        operator (str): The comparison operator, one of '>', '>=', '<', '<=', '='.
        value (Union[date, str]): The date value, which can be a `date` object or a string in 'YYYY-MM-DD' format.
        date_format (str): The format to use for parsing date strings (default is 'YYYY-MM-DD').
    """

    def __init__(
        self, operator: str, value: Union[date, str], date_format="YYYY-MM-DD"
    ):
        if isinstance(value, str):
            value = datetime.strptime(value, date_format).date()
        super(Date, self).__init__(operator, value)


class Before(Date):
    """
    Represents a threshold where a date must be strictly before the specified value.
    """

    def __init__(self, value: Union[date, str], **kwargs):
        super(Before, self).__init__("<", value)


class BeforeOrOn(Date):
    """
    Represents a threshold where a date must be on or before the specified value.
    """

    def __init__(self, value: Union[date, str], **kwargs):
        super(BeforeOrOn, self).__init__("<=", value)


class After(Date):
    """
    Represents a threshold where a date must be strictly after the specified value.
    """

    def __init__(self, value: Union[date, str], **kwargs):
        super(After, self).__init__(">", value)


class AfterOrOn(Date):
    """
    Represents a threshold where a date must be on or after the specified value.
    """

    def __init__(self, value: Union[date, str], **kwargs):
        super(AfterOrOn, self).__init__(">=", value)


def DateFilter(
    min_date: Optional[Union[Date, After, AfterOrOn]] = None,
    max_date: Optional[Union[Date, Before, BeforeOrOn]] = None,
    column_name: str = "EVENT_DATE",
):
    """
    DateFilter is a specialized ValueFilter for handling date-based filtering.

    Parameters:
        min_date: The minimum date condition. Recommended to pass either After or AfterOrOn.
        max_date: The maximum date condition. Recommended to pass either Before or BeforeOrOn.
        column_name: The name of the column to apply the filter on. Defaults to "EVENT_DATE".
    """
    # For some reason, implementing DateFilter as a subclass of ValueFilter messes up the serialization. So instead we implement DateFilter as a function that looks like a class and just returns a ValueFilter instance.
    return ValueFilter(min_value=min_date, max_value=max_date, column_name=column_name)



from typing import Union
from phenex.phenotypes.phenotype import Phenotype


class WithinSameEncounterPhenotype(Phenotype):
    """
    WithinSameEncounterPhenotype is a phenotype that filters a target phenotype based on the occurrence of an anchor phenotype within the same encounter. This phenotype can only be used with CodelistPhenotypes and MeasurementPhenotypes as anchor/phenotype.

    Parameters:
        name: The name of the phenotype. Optional. If not passed, name will be derived from the name of the codelist.
        anchor_phenotype: The phenotype which should define the encounter of interest
        phenotype: The phenotype which should occur within the same encounter of interest, defined by the anchor_phenotype
        column_name: The name of the column which must be shared between anchor_phenotype and phenotype


    Attributes:
        table (PhenotypeTable): The resulting phenotype table after filtering (None until execute is called)

    Examples:

    Example: Imaging procedure within the same encounter as Pulmonary Embolism (OMOP)
        ```python
        from phenex.phenotypes import CodelistPhenotype, WithinSameEncounterPhenotype
        from phenex.codelists import Codelist
        from phenex.mappers import OMOPDomains
        from phenex.filters import DateFilter, CategoricalFilter, Value
        from phenex.ibis_connect import SnowflakeConnector

        pe_diagnosis = CodelistPhenotype(
            name = "diagnosis_of_PE",
            domain = "CONDITION_OCCURRENCE",
            codelist = Codelist(codelist={'SNOMED': [608954, 4120091, 45768439, 45768888, 37165995, 436768, 44782732, 45768887, 45771016, 4017134, 4240832, 4129841, 43530934, 4219469, 442055, 442051, 442050, 433545, 437060, 435026, 433832, 440477, 439380, 439379, 4331168, 4108681, 37160752, 4091708, 1244882, 440417, 37109911, 3655209, 3655210, 45757145, 37016922, 43530605, 4119608, 4253796, 45766471, 4121618, 4236271, 36713113, 35615055, 40479606, 4119607, 4119609]}).copy(use_code_type=False),
            return_date = "first",
        )

        # where column id alone is idnetical left join
        imaging = CodelistPhenotype(
            name = 'imaging',
            domain = 'PROCEDURE_OCCURRENCE',
            codelist = Codelist(codelist={'CPT4': [2211381, 2211379, 2211380, 2212002, 2212004, 2212003, 2212011, 42742555, 42742554, 2212006, 2212005, 2212008, 2212010, 2212009, 42742557, 42742556]}).copy(use_code_type=False),
        )


        imaging_in_entry_hospitalization = WithinSameEncounterPhenotype(
            name='imaging_in_entry_hospitalization',
            anchor_phenotype = pe_diagnosis,
            phenotype = imaging,
            column_name = 'VISIT_OCCURRENCE_ID'
        )

        ```
    """

    def __init__(
        self,
        anchor_phenotype: Union["CodelistPhenotype", "MeasurementPhenotype"],
        phenotype: Union["CodelistPhenotype", "MeasurementPhenotype"],
        column_name: str,
        **kwargs,
    ):
        super(WithinSameEncounterPhenotype, self).__init__(**kwargs)
        if (
            anchor_phenotype.__class__.__name__
            not in ["CodelistPhenotype", "MeasurementPhenotype"]
        ) or (
            phenotype.__class__.__name__
            not in ["CodelistPhenotype", "MeasurementPhenotype"]
        ):
            raise ValueError(
                "Both anchor_phenotype and phenotype must be of type CodelistPhenotype or MeasurementPhenotype"
            )

        self.anchor_phenotype = anchor_phenotype
        self.phenotype = phenotype
        self.column_name = column_name
        self.children.append(self.anchor_phenotype)

    def _execute(self, tables) -> "PhenotypeTable":
        # Subset the raw anchor data that occurs on the same day as the anchor date in order to get the column of interest
        _anchor_table = tables[self.anchor_phenotype.domain]
        _anchor_table = _anchor_table.join(
            self.anchor_phenotype.table,
            (self.anchor_phenotype.table.PERSON_ID == _anchor_table.PERSON_ID)
            & (self.anchor_phenotype.table.EVENT_DATE == _anchor_table.EVENT_DATE),
            how="inner",
        ).select(["PERSON_ID", self.column_name])

        # Subset the target phenotype raw data for patient id and column name of interest
        _table = tables[self.phenotype.domain]
        _table = _table.join(
            _anchor_table,
            (_anchor_table.PERSON_ID == _table.PERSON_ID)
            & (_anchor_table[self.column_name] == _table[self.column_name]),
            how="inner",
        )

        # run the target phenotype on the subsetted data
        return self.phenotype._execute({self.phenotype.domain: _table})



import json as pyjson
from .to_dict import to_dict
from .from_dict import from_dict


def dump(obj, fp, **kwargs):
    """
    Serialize PhenEx objects as a JSON formatted stream to `fp` (a `.write()`-supporting file-like object).

    Example:
    ```python
    from phenex.codelists import Codelist
    from phenex.phenotypes import CodelistPhenotype
    import phenex.util.serialization.json as pxjson


    # Initialize with a list
    cl = Codelist(
        ['x', 'y', 'z'],
        'mycodelist'
    )
    entry = CodelistPhenotype(
        return_date="first",
        codelist=cl,
        domain="DRUG_EXPOSURE",
        date_range=study_period,
    )

    cohort = Cohort(
        entry=entry
    )

    PATH = ''

    # serialize the phenex object to file
    with open(PATH, "w") as f:
        pxjson.dump(cohort, f, indent=4)
    ```

    """
    pyjson.dump(obj.to_dict(), fp, **kwargs)


def dumps(obj, **kwargs):
    """
    Serialize a PhenEx object to a JSON formatted `str`.
    """
    return pyjson.dumps(to_dict(obj), **kwargs)


def load(fp, **kwargs):
    """
    Deserialize `fp` (a `.read()`-supporting file-like object containing a JSON document) to a PhenEx object.

    Example:
    ```python
    PATH = ''
    # serialize the phenex object to file
    with open(PATH, "r") as f:
        cohort = pxjson.load(f)
    ```
    """
    data = pyjson.load(fp, **kwargs)
    return from_dict(data)


def loads(s, **kwargs):
    """
    Deserialize `s` (a `str`, `bytes` or `bytearray` instance containing a JSON document) to a PhenEx object.
    """
    data = pyjson.loads(s, **kwargs)
    return from_dict(data)



from typing import Union, List, Dict
from datetime import date
from phenex.phenotypes.phenotype import Phenotype
from phenex.filters.relative_time_range_filter import RelativeTimeRangeFilter
from phenex.filters.date_filter import DateFilter
from phenex.filters.filter import AndFilter, OrFilter
from phenex.aggregators import First, Last
from phenex.filters.categorical_filter import CategoricalFilter
from phenex.tables import PHENOTYPE_TABLE_COLUMNS, PhenotypeTable
from phenex.phenotypes.functions import select_phenotype_columns
from ibis import _
import ibis


def check_categorical_filters_share_same_domain(filter, domain):
    # if any leaf node of the tree has a different domain, return false
    if isinstance(filter, AndFilter) or isinstance(filter, OrFilter):
        filter1_same_domain = check_categorical_filters_share_same_domain(
            filter.filter1, domain
        )
        filter2_same_domain = check_categorical_filters_share_same_domain(
            filter.filter2, domain
        )
        return filter1_same_domain and filter2_same_domain
    # else check if the current leaf node has the same domain
    else:
        if filter.domain is not None:
            if filter.domain != domain:
                return False
    return True


class CategoricalPhenotype(Phenotype):
    """
    CategoricalPhenotype calculates phenotype whose VALUE is discrete, such for sex, race, or ethnicity. Categorical Phenotype is especially helpful as a baseline characteristic from PERSON like tables.
    The returned Phenotype has the following interpretation:

    DATE: If when='before', then DATE is the beginning of the coverage period containing the anchor phenotype. If when='after', then DATE is the end of the coverage period containing the anchor date.
    VALUE: Coverage (in days) relative to the anchor date. By convention, always non-negative.


    Parameters:
        name: Name of the phenotype.
        domain: Domain of the phenotype.
        categorical_filter: Use CategoricalFilter to input allowed values for the categorical variable. If not passed, all values are returned.
    """

    def __init__(
        self,
        domain: str,
        categorical_filter: CategoricalFilter,
        date_range: DateFilter = None,
        relative_time_range: Union[
            RelativeTimeRangeFilter, List[RelativeTimeRangeFilter]
        ] = None,
        return_date=None,
        **kwargs,
    ):
        super(CategoricalPhenotype, self).__init__(**kwargs)
        self.domain = domain
        if not check_categorical_filters_share_same_domain(
            categorical_filter, self.domain
        ):
            raise ValueError("CategoricalPhenotype only works on a single domain.")
        self.categorical_filter = categorical_filter
        self.date_range = date_range
        self.return_date = return_date
        assert self.return_date in [
            "first",
            "last",
            "nearest",
            "all",
            None,
        ], f"Unknown return_date: {return_date}"

        if isinstance(relative_time_range, RelativeTimeRangeFilter):
            relative_time_range = [relative_time_range]

        self.relative_time_range = relative_time_range
        if self.relative_time_range is not None:
            for rtr in self.relative_time_range:
                if rtr.anchor_phenotype is not None:
                    self.children.append(rtr.anchor_phenotype)

    def _execute(self, tables) -> PhenotypeTable:
        table = tables[self.domain]
        table = self._perform_categorical_filtering(table)
        table = self._perform_time_filtering(table)
        table = self._perform_date_selection(table)

        if isinstance(self.categorical_filter, CategoricalFilter):
            table = table.mutate(
                VALUE=table[self.categorical_filter.column_name],
                EVENT_DATE=ibis.null(date),
            )
        return select_phenotype_columns(table)

    def _perform_categorical_filtering(self, table):
        table = self.categorical_filter.filter(table)
        return table

    def _perform_time_filtering(self, table):
        if self.date_range is not None:
            table = self.date_range.filter(table)
        if self.relative_time_range is not None:
            for rtr in self.relative_time_range:
                table = rtr.filter(table)
        return table

    def _perform_date_selection(self, table):
        if self.return_date is None or self.return_date == "all":
            return table
        if self.return_date == "first":
            aggregator = First()
        elif self.return_date == "last":
            aggregator = Last()
        else:
            raise ValueError(f"Unknown return_date: {self.return_date}")
        return aggregator.aggregate(table)



from typing import Dict
from ibis.expr.types.relations import Table

from phenex.tables import *


class DomainsDictionary:
    """
    A DomainsDictionary is used to map an entire database from an arbitrary schema to a PhenEx internal representation.

    Attributes:
        domains_dict (Dict[str, class]): A dictionary where keys are domain names and values are uninstantiated PhenexTable class objects.

    Methods:
        get_mapped_tables(con, database=None) -> Dict[str, Table]:
            Get all tables mapped to PhenEx representation using the given connection.
        set_mapped_tables(con, con, overwrite=False) -> None:
            Create a view for all mapped tables in the destination database.
    """

    def __init__(self, domains_dict):
        self.domains_dict = domains_dict

    def set_mapped_tables(self, con, overwrite=False) -> Dict[str, Table]:
        """
        Get all tables mapped to PhenEx representation using the given connection.

        If a database is not provided, the current database of the connection is used to find the tables.

        Args:
            con: The connection to the database.
            database (Optional[str]): The name of the database. Defaults to the current database of the connection.

        Returns:
            Dict[str, Table]: A dictionary where keys are domain names and values are mapped tables.
        """
        existing_tables = con.dest_connection.list_tables(
            database=con.SNOWFLAKE_DEST_DATABASE
        )
        for domain, mapper in self.domains_dict.items():
            if domain not in existing_tables or overwrite:
                t = con.get_source_table(mapper.NAME_TABLE)
                mapped_table = mapper(t).table
                con.create_view(
                    mapped_table, name_table=mapper.NAME_TABLE, overwrite=overwrite
                )

    def get_mapped_tables(self, con) -> Dict[str, PhenexTable]:
        """
        Get all tables mapped to PhenEx representation using the given connection.

        If a database is not provided, the current database of the connection is used to find the tables.

        Args:
            con: The connection to the database.
            database (Optional[str]): The name of the database. Defaults to the current database of the connection.

        Returns:
            Dict[str, PhenexTable]: A dictionary where keys are domain names and values are mapped tables.
        """
        # self.set_mapped_tables(con)
        mapped_tables = {}
        for domain, mapper in self.domains_dict.items():
            mapped_tables[domain] = mapper(con.get_source_table(mapper.NAME_TABLE))
        return mapped_tables


#
# OMOP Column Mappers
#
class OMOPPersonTable(PhenexPersonTable):
    NAME_TABLE = "PERSON"
    DEFAULT_MAPPING = {"PERSON_ID": "PERSON_ID", "DATE_OF_BIRTH": "BIRTH_DATETIME"}
    JOIN_KEYS = {
        "OMOPConditionOccurenceTable": ["PERSON_ID"],
        "OMOPVisitOccurrenceTable": ["PERSON_ID"],
    }


class OMOPVisitOccurrenceTable(PhenexObservationPeriodTable):
    NAME_TABLE = "VISIT_OCCURRENCE"
    JOIN_KEYS = {
        "OMOPPersonTable": ["PERSON_ID"],
        "OMOPConditionOccurenceTable": ["PERSON_ID", "VISIT_OCCURRENCE_ID"],
        "OMOPVisitDetailTable": ["PERSON_ID", "VISIT_OCCURRENCE_ID"],
    }
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "START_DATE": "VISIT_START_DATE",
        "END_DATE": "VISIT_END_DATE",
    }


class OMOPVisitDetailTable(PhenexTable):
    NAME_TABLE = "VISIT_DETAIL"
    RELATIONSHIPS = {
        "OMOPPersonTable": ["PERSON_ID"],
        "OMOPVisitOccurrenceTable": ["PERSON_ID", "VISIT_OCCURRENCE_ID"],
    }


class OMOPConditionOccurenceTable(CodeTable):
    NAME_TABLE = "CONDITION_OCCURRENCE"
    JOIN_KEYS = {
        "OMOPPersonTable": ["PERSON_ID"],
        "OMOPVisitOccurrenceTable": [
            "PERSON_ID",
            "VISIT_OCCURRENCE_ID",
        ],
    }
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "EVENT_DATE": "CONDITION_START_DATE",
        "CODE": "CONDITION_CONCEPT_ID",
    }


class OMOPObservationTable(MeasurementTable):
    NAME_TABLE = "OBSERVATION"
    JOIN_KEYS = {
        "OMOPPersonTable": ["PERSON_ID"],
        "OMOPVisitOccurrenceTable": [
            "PERSON_ID",
            "VISIT_OCCURRENCE_ID",
        ],
    }
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "EVENT_DATE": "OBSERVATION_DATE",
        "CODE": "OBSERVATION_CONCEPT_ID",
        "VALUE": "VALUE_AS_NUMBER",
    }


class OMOPMeasurementTable(MeasurementTable):
    NAME_TABLE = "MEASUREMENT"
    JOIN_KEYS = {
        "OMOPPersonTable": ["PERSON_ID"],
        "OMOPVisitOccurrenceTable": [
            "PERSON_ID",
            "VISIT_OCCURRENCE_ID",
        ],
        "OMOPVisitDetailTable": [
            "PERSON_ID",
            "VISIT_DETAIL_ID",
        ],
    }
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "EVENT_DATE": "MEASUREMENT_DATE",
        "CODE": "MEASUREMENT_TYPE_CONCEPT_ID",
        "VALUE": "VALUE_AS_NUMBER",
    }


class OMOPDeathTable(PhenexTable):
    NAME_TABLE = "DEATH"
    JOIN_KEYS = {"OMOPPersonTable": ["PERSON_ID"]}
    KNOWN_FIELDS = ["PERSON_ID", "DATE_OF_DEATH"]
    DEFAULT_MAPPING = {"PERSON_ID": "PERSON_ID", "DATE_OF_DEATH": "DEATH_DATE"}


class OMOPProcedureOccurrenceTable(CodeTable):
    NAME_TABLE = "PROCEDURE_OCCURRENCE"
    JOIN_KEYS = {
        "OMOPPersonTable": ["PERSON_ID"],
        "OMOPVisitOccurrenceTable": ["PERSON_ID", "VISIT_OCCURRENCE_ID"],
    }
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "EVENT_DATE": "PROCEDURE_DATE",
        "CODE": "PROCEDURE_CONCEPT_ID",
    }


class OMOPDrugExposureTable(CodeTable):
    NAME_TABLE = "DRUG_EXPOSURE"
    JOIN_KEYS = {
        "OMOPPersonTable": ["PERSON_ID"],
        "OMOPVisitOccurrenceTable": ["PERSON_ID", "VISIT_OCCURRENCE_ID"],
    }
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "EVENT_DATE": "DRUG_EXPOSURE_START_DATE",
        "CODE": "DRUG_CONCEPT_ID",
    }


class OMOPConditionOccurrenceSourceTable(CodeTable):
    NAME_TABLE = "CONDITION_OCCURRENCE"
    JOIN_KEYS = {
        "OMOPPersonTable": ["PERSON_ID"],
        "OMOPVisitOccurrenceTable": ["PERSON_ID", "VISIT_OCCURRENCE_ID"],
    }
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "EVENT_DATE": "CONDITION_START_DATE",
        "CODE": "CONDITION_SOURCE_VALUE",
    }


class OMOPProcedureOccurrenceSourceTable(CodeTable):
    NAME_TABLE = "PROCEDURE_OCCURRENCE"
    JOIN_KEYS = {
        "OMOPPersonTable": ["PERSON_ID"],
        "OMOPVisitOccurrenceTable": ["PERSON_ID", "VISIT_OCCURRENCE_ID"],
    }
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "EVENT_DATE": "PROCEDURE_DATE",
        "CODE": "PROCEDURE_SOURCE_VALUE",
    }


class OMOPDrugExposureSourceTable(CodeTable):
    NAME_TABLE = "DRUG_EXPOSURE"
    JOIN_KEYS = {
        "OMOPPersonTable": ["PERSON_ID"],
        "OMOPVisitOccurrenceTable": ["PERSON_ID", "VISIT_OCCURRENCE_ID"],
    }
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "EVENT_DATE": "DRUG_EXPOSURE_START_DATE",
        "CODE": "DRUG_SOURCE_VALUE",
    }


class OMOPPersonTableSource(PhenexPersonTable):
    NAME_TABLE = "PERSON"
    JOIN_KEYS = {
        "OMOPConditionOccurenceTable": ["PERSON_ID"],
        "OMOPVisitOccurrenceTable": ["PERSON_ID"],
    }
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "DATE_OF_BIRTH": "BIRTH_DATETIME",
        "YEAR_OF_BIRTH": "YEAR_OF_BIRTH",
        "SEX": "GENDER_SOURCE_VALUE",
        "ETHNICITY": "ETHNICITY_SOURCE_VALUE",
    }


class OMOPObservationPeriodTable(PhenexObservationPeriodTable):
    NAME_TABLE = "OBSERVATION_PERIOD"
    JOIN_KEYS = {"OMOPPersonTable": ["PERSON_ID"]}
    DEFAULT_MAPPING = {
        "PERSON_ID": "PERSON_ID",
        "START_DATE": "OBSERVATION_PERIOD_START_DATE",
        "END_DATE": "OBSERVATION_PERIOD_END_DATE",
    }


class OMOPConceptTable(PhenexTable):
    NAME_TABLE = "CONCEPT"
    JOIN_KEYS = {"OMOPConditionOccurenceTable": ["CONCEPT_ID", "CONDITION_CONCEPT_ID"]}
    KNOWN_FIELDS = ["CONCEPT_ID", "CONCEPT_NAME"]
    DEFAULT_MAPPING = {}


#
# Domains
#
OMOPs = {
    "PERSON": OMOPPersonTable,
    "VISIT_OCCURRENCE": OMOPVisitOccurrenceTable,
    "VISIT_DETAIL": OMOPVisitDetailTable,
    "CONDITION_OCCURRENCE": OMOPConditionOccurenceTable,
    "DEATH": OMOPDeathTable,
    "PROCEDURE_OCCURRENCE": OMOPProcedureOccurrenceTable,
    "DRUG_EXPOSURE": OMOPDrugExposureTable,
    "CONDITION_OCCURRENCE_SOURCE": OMOPConditionOccurrenceSourceTable,
    "PROCEDURE_OCCURRENCE_SOURCE": OMOPProcedureOccurrenceSourceTable,
    "DRUG_EXPOSURE_SOURCE": OMOPDrugExposureSourceTable,
    "PERSON_SOURCE": OMOPPersonTableSource,
    "OBSERVATION_PERIOD": OMOPObservationPeriodTable,
    "OBSERVATION": OMOPObservationTable,
    "MEASUREMENT": OMOPMeasurementTable,
}
OMOPDomains = DomainsDictionary(OMOPs)



import datetime, os
import pandas as pd
import ibis
from phenex.test.cohort_test_generator import CohortTestGenerator
from phenex.codelists import Codelist
from phenex.phenotypes import (
    AgePhenotype,
    CategoricalPhenotype,
    CodelistPhenotype,
    Cohort,
    TimeRangePhenotype,
    SexPhenotype,
)
from phenex.filters import *
from phenex.tables import PhenexPersonTable, CodeTable, PhenexObservationPeriodTable
from phenex.test.cohort.test_mappings import (
    PersonTableForTests,
    DrugExposureTableForTests,
    ObservationPeriodTableForTests,
    ConditionOccurenceTableForTests,
)


def create_cohort():
    """
    ## DEMO STUDY DEFINITION
    INCLUSION CRITERIA
        * Female patient
        * Aged ≥18 as of the index date
        * Registered in CPRD at least 365 days prior to the index date
        *  At least one drug exposure to ET, defined as a prescription for tamoxifen, aromatase inhibitors (AIs), or gonadotropin releasing hormone (GnRH) antagonists  between 1 January 2010 through 31 December 2020.
    The date of the first ET prescription in this time period is the index date.
    AIs will include: aminoglutethimide, formestane, anastrozole, letrozole, vorozole, exemestane
    GnRH agonists will include: buserelin, leuprorelin/leuprolide, goserelin, triptorelin, histrelin, leuprorelin and bicalutamide, gonadorelin, nafarelin, ganirelix, cetrorelix, elagolix, linzagolix, relugolix, estradiol and norethisterone, degarelix, abarelix
        * BC diagnosis at any time prior to ET prescription in the baseline period
        * Permanently registered acceptable patient

    EXCLUSION CRITERIA
    Any prescription for tamoxifen or AIs before the BC diagnosis date (prevalent users)
    """
    study_period = DateFilter(
        min_date=AfterOrOn(datetime.date(2010, 1, 1)),
        max_date=BeforeOrOn(datetime.date(2020, 12, 31)),
    )

    entry = CodelistPhenotype(
        return_date="first",
        codelist=Codelist(["d1"]).copy(use_code_type=False),
        domain="DRUG_EXPOSURE",
        date_range=study_period,
    )

    inclusion, exclusion = define_inclusion_exclusion_criteria(entry)

    cohort = Cohort(
        name="bc",
        entry_criterion=entry,
        inclusions=inclusion,
        exclusions=exclusion,
    )

    return cohort


def define_inclusion_exclusion_criteria(entry):
    continuous_coverage = TimeRangePhenotype(
        relative_time_range=RelativeTimeRangeFilter(
            min_days=GreaterThanOrEqualTo(365), anchor_phenotype=entry
        )
    )

    age_18 = AgePhenotype(
        value_filter=ValueFilter(min_value=GreaterThanOrEqualTo(18)),
        anchor_phenotype=entry,
    )

    sex = SexPhenotype(categorical_filter=CategoricalFilter(allowed_values=[2]))

    quality = CategoricalPhenotype(
        domain="PERSON",
        name="data_quality",
        categorical_filter=CategoricalFilter(
            allowed_values=[1],
            column_name="ACCEPTABLE",
        ),
    )

    breast_cancer = CodelistPhenotype(
        name="breast_cancer",
        codelist=Codelist(["b1"]).copy(use_code_type=False),
        domain="CONDITION_OCCURRENCE",
        return_date="first",
        relative_time_range=RelativeTimeRangeFilter(
            when="before", min_days=GreaterThan(0), anchor_phenotype=entry
        ),
    )

    inclusion_criteria = [breast_cancer, continuous_coverage, quality, age_18, sex]

    tamoxifen = CodelistPhenotype(
        name="prior_et_usage",
        codelist=Codelist(["d4", "d5", "d6"]).copy(use_code_type=False),
        domain="DRUG_EXPOSURE",
        relative_time_range=RelativeTimeRangeFilter(
            anchor_phenotype=breast_cancer,
            when="before",
            min_days=GreaterThan(0),
        ),
    )

    exclusion_criteria = []  # [tamoxifen]

    return inclusion_criteria, exclusion_criteria


class SimpleCohortTestGenerator(CohortTestGenerator):
    def define_cohort(self):
        return create_cohort()

    def define_mapped_tables(self):
        self.con = ibis.duckdb.connect()
        df_allvalues = create_test_data()

        # create dummy person table
        df_person = pd.DataFrame(df_allvalues[["PATID", "YOB", "GENDER", "ACCEPTABLE"]])
        schema_person = {"PATID": str, "YOB": int, "GENDER": int, "ACCEPTABLE": int}
        person_table = PersonTableForTests(
            self.con.create_table("PERSON", df_person, schema=schema_person)
        )

        # create dummy observation period table
        df_observation = pd.DataFrame(df_allvalues[["PATID", "REGSTARTDATE"]])
        df_observation["REGENDDATE"] = datetime.date(2020, 10, 10)
        schema_observation = {
            "PATID": str,
            "REGSTARTDATE": datetime.date,
            "REGENDDATE": datetime.date,
        }
        observation_period_table = ObservationPeriodTableForTests(
            self.con.create_table(
                "OBSERVATION_PERIOD", df_observation, schema=schema_observation
            )
        )

        # create dummy condition occurrence table
        df_condition_occurrence = pd.DataFrame(
            df_allvalues[["PATID", "breast_cancer_code", "breast_cancer_date"]]
        )
        df_condition_occurrence.columns = ["PATID", "MEDCODEID", "OBSDATE"]
        schema_condition_occurrence = {
            "PATID": str,
            "MEDCODEID": str,
            "OBSDATE": datetime.date,
        }
        condition_occurrence_table = ConditionOccurenceTableForTests(
            self.con.create_table(
                "CONDITION_OCCURRENCE",
                df_condition_occurrence,
                schema=schema_condition_occurrence,
            )
        )

        # create drug exposure table
        df_drug_exposure = pd.DataFrame(df_allvalues[["PATID", "entry", "entry_date"]])
        df_drug_exposure.columns = ["PATID", "PRODCODEID", "ISSUEDATE"]
        schema_drug_exposure = {
            "PATID": str,
            "PRODCODEID": str,
            "ISSUEDATE": datetime.date,
        }
        drug_exposure_table = DrugExposureTableForTests(
            self.con.create_table(
                "DRUG_EXPOSURE", df_drug_exposure, schema=schema_drug_exposure
            )
        )

        return {
            "PERSON": person_table,
            "CONDITION_OCCURRENCE": condition_occurrence_table,
            "DRUG_EXPOSURE": drug_exposure_table,
            "OBSERVATION_PERIOD": observation_period_table,
        }

    def define_expected_output(self):
        df_counts_inclusion = pd.DataFrame()
        df_counts_inclusion["phenotype"] = [
            x.upper()
            for x in [
                "breast_cancer",
                "time_range",
                "data_quality",
                "age",
                "sex",
            ]
        ]
        df_counts_inclusion["n"] = [16, 32, 32, 32, 32]

        df_counts_exclusion = pd.DataFrame()
        df_counts_exclusion["phenotype"] = ["prior_et_usage"]
        df_counts_exclusion["n"] = [0]

        test_infos = {
            "counts_inclusion": df_counts_inclusion,
            "counts_exclusion": df_counts_exclusion,
        }
        return test_infos


def create_test_data():
    # this is a list of all the parameters i want every possible combination of
    inclusion = [
        {
            "name": "entry",
            "values": ["d1", "d4"],
        },  # the first value is the allowed value, the second is the not allowed value
        {
            "name": "entry_date",
            "values": [datetime.date(2020, 1, 1), datetime.date(2000, 1, 1)],
        },  # first date within study period, second date after
        {"name": "breast_cancer_code", "values": ["b1", "b4"]},
        {
            "name": "breast_cancer_date",
            "values": [datetime.date(2019, 5, 1), datetime.date(2021, 10, 1)],
        },
        {
            "name": "REGSTARTDATE",
            "values": [datetime.date(2018, 1, 1), datetime.date(2019, 10, 1)],
        },  # this is the start date of coverage. i will manually sett index date (date of entry) as 2020.01.01. cc callculated from there. end date similarly a date after index date
        {"name": "ACCEPTABLE", "values": [1, 0]},
        {"name": "YOB", "values": [1970, 2010]},
        {"name": "GENDER", "values": [2, 1]},
        # {"name":"prior_et_use", "values":['d7','d5']},
    ]

    # create the dataframe with two rows; first patient fulfills entry criteria, second does not
    item = inclusion[0]
    df = pd.DataFrame()
    df[item["name"]] = item["values"]

    # iterate over each following criteria, duplicating the previous values
    for item in inclusion[1:]:
        _dfs = []
        for value in item["values"]:
            _df = pd.DataFrame(df)
            _df[item["name"]] = value
            _dfs.append(_df)
        df = pd.concat(_dfs)
    # create appropriate patient ids. only patient 0 fulfills all criteria!
    df["PATID"] = [f"P{i}" for i in range(df.shape[0])]
    return df


def test_simple_cohort():
    cprd_study = SimpleCohortTestGenerator()
    cprd_study.run_tests()


if __name__ == "__main__":
    test_simple_cohort()



from datetime import date
from ibis.expr.types.relations import Table
import ibis
from phenex.util.serialization.to_dict import to_dict


class VerticalDateAggregator:
    def __init__(
        self,
        aggregation_index=["PERSON_ID"],
        aggregation_function="sum",
        event_date_column="EVENT_DATE",
        reduce=False,
    ):
        self.aggregation_index = aggregation_index
        self.aggregation_function = aggregation_function
        self.event_date_column = event_date_column
        self.reduce = reduce

    def aggregate(self, input_table: Table):
        # Define the window specification
        partition_cols = [
            getattr(input_table, col) if isinstance(col, str) else col
            for col in self.aggregation_index
        ]
        window_spec = ibis.window(
            group_by=partition_cols, order_by=input_table[self.event_date_column]
        )

        # Apply the specified aggregation function to the event_date_column
        event_date_col = getattr(input_table, self.event_date_column)
        if self.aggregation_function == "max":
            aggregated_date = event_date_col.max().over(window_spec)
        elif self.aggregation_function == "min":
            aggregated_date = event_date_col.min().over(window_spec)
        else:
            raise ValueError(
                f"Unsupported aggregation function: {self.aggregation_function}"
            )

        # Add the aggregated date as a new column
        input_table = input_table.mutate(aggregated_date=aggregated_date)

        # Filter rows where the original date matches the aggregated date
        input_table = input_table.filter(
            input_table[self.event_date_column] == input_table.aggregated_date
        )

        # Select the necessary columns

        # Apply the distinct reduction if required
        if self.reduce:
            selected_columns = self.aggregation_index + [self.event_date_column]
            input_table = input_table.select(selected_columns).distinct()
            input_table = input_table.mutate(VALUE=ibis.null().cast("int32"))

        return input_table

    def to_dict(self):
        return to_dict(self)


class Nearest(VerticalDateAggregator):
    def __init__(self, **kwargs):
        super().__init__(aggregation_function="max", **kwargs)


class First(VerticalDateAggregator):
    def __init__(self, **kwargs):
        super().__init__(aggregation_function="min", **kwargs)


class Last(VerticalDateAggregator):
    def __init__(self, **kwargs):
        super().__init__(aggregation_function="max", **kwargs)


class ValueAggregator:
    def __init__(
        self,
        aggregation_column=None,
        aggregation_function="min",
        aggregation_index=["PERSON_ID"],
        reduce=True,
    ):
        # allowed values for aggregation_function are any valid aggregation
        # function in SQL (except some make no sense for date, e.g variance)
        self.aggregation_column = aggregation_column
        self.aggregation_function = aggregation_function
        self.aggregation_index = aggregation_index

        self.reduce = reduce
        # if true, max one row per unique combination of index columns
        # otherwise, row count is preserved

    def aggregate(self, input_table: Table):
        # Get the aggregation index columns
        _aggregation_index_cols = [
            getattr(input_table, col) for col in self.aggregation_index
        ]

        # Determine the aggregation column
        if self.aggregation_column is None:
            _aggregation_column = input_table.VALUE
        else:
            _aggregation_column = getattr(input_table, self.aggregation_column)

        ibis.options.interactive = True
        # Define the window specification
        window_spec = ibis.window(group_by=_aggregation_index_cols)

        # Function to apply based on aggregation_function
        if self.aggregation_function == "median":
            aggregation = _aggregation_column.median().over(window_spec)
        elif self.aggregation_function == "mean":
            aggregation = _aggregation_column.mean().over(window_spec)
        elif self.aggregation_function == "max":
            aggregation = _aggregation_column.max().over(window_spec)
        elif self.aggregation_function == "min":
            aggregation = _aggregation_column.min().over(window_spec)
        else:
            raise ValueError(
                f"Unsupported aggregation function: {self.aggregation_function}"
            )
        # Select the necessary columns
        selected_columns = _aggregation_index_cols + [aggregation.name("VALUE")]

        # Apply the distinct reduction if required
        if self.reduce:
            input_table = input_table.select(selected_columns).distinct()
            input_table = input_table.mutate(EVENT_DATE=ibis.null(date))
            return input_table
        else:
            return input_table.select(selected_columns)

    def to_dict(self):
        return to_dict(self)


class Mean(ValueAggregator):
    def __init__(self, **kwargs):
        super(Mean, self).__init__(aggregation_function="mean", **kwargs)


class Max(ValueAggregator):
    def __init__(self, **kwargs):
        super(Max, self).__init__(aggregation_function="max", **kwargs)


class Min(ValueAggregator):
    def __init__(self, **kwargs):
        super(Min, self).__init__(aggregation_function="min", **kwargs)


class DailyValueAggregator(ValueAggregator):
    def __init__(self, aggregation_index=["PERSON_ID", "EVENT_DATE"], **kwargs):
        super(DailyValueAggregator, self).__init__(
            aggregation_index=aggregation_index, **kwargs
        )


class DailyMean(DailyValueAggregator):
    def __init__(self):
        super(DailyMean, self).__init__(aggregation_function="mean")


class DailyMedian(DailyValueAggregator):
    def __init__(self):
        super(DailyMedian, self).__init__(aggregation_function="median")


class DailyMax(DailyValueAggregator):
    def __init__(self, **kwargs):
        super(DailyMax, self).__init__(aggregation_function="max", **kwargs)


class DailyMin(DailyValueAggregator):
    def __init__(self, **kwargs):
        super(DailyMin, self).__init__(aggregation_function="min", **kwargs)



from typing import Optional, List
import os
import ibis
from ibis.backends import BaseBackend


# Snowflake connection function
def _check_env_vars(*vars: str) -> None:
    """
    Check if the required environment variables are set.

    Args:
        *vars: Variable length argument list of environment variable names.

    Raises:
        EnvironmentError: If any of the required environment variables are missing.
    """
    missing_vars = [var for var in vars if os.getenv(var) is None]
    if missing_vars:
        raise EnvironmentError(
            f"Missing required environment variables: {', '.join(missing_vars)}. Add to .env file or set in the environment."
        )


class SnowflakeConnector:
    """
    SnowflakeConnector manages input (read) and output (write) connections to Snowflake using Ibis. Parameters may be specified with environment variables of the same name or through the __init__() method interface. Variables passed through __init__() take precedence.

    Attributes:
        SNOWFLAKE_USER: Snowflake user name.
        SNOWFLAKE_ACCOUNT: Snowflake account identifier.
        SNOWFLAKE_WAREHOUSE: Snowflake warehouse name.
        SNOWFLAKE_ROLE: Snowflake role name.
        SNOWFLAKE_PASSWORD: Snowflake password. If not specified, will attempt to authenticate with externalbrowser.
        SNOWFLAKE_SOURCE_DATABASE: Snowflake source database name. Use a fully qualified database name (in snowflake terminology DATABASE.SCHEMA; ibis calls this a "database").
        SNOWFLAKE_DEST_DATABASE: Snowflake destination database name. Use a fully qualified database name (in snowflake terminology DATABASE.SCHEMA; ibis calls this a "database").

    Methods:
        connect_dest() -> BaseBackend:
            Establishes and returns an Ibis backend connection to the destination Snowflake database and schema.

        connect_source() -> BaseBackend:
            Establishes and returns an Ibis backend connection to the source Snowflake database and schema.

        get_source_table(name_table: str) -> Table:
            Retrieves a table from the source Snowflake database.

        get_dest_table(name_table: str) -> Table:
            Retrieves a table from the destination Snowflake database.

        create_view(table: Table, name_table: Optional[str] = None, overwrite: bool = False) -> View:
            Create a view of a table in the destination Snowflake database.

        create_table(table: Table, name_table: Optional[str] = None, overwrite: bool = False) -> Table:
            Materialize a table in the destination Snowflake database.

        drop_table(name_table: str) -> None:
            Drop a table from the destination Snowflake database.

        drop_view(name_table: str) -> None:
            Drop a view from the destination Snowflake database.
    """

    def __init__(
        self,
        SNOWFLAKE_USER: Optional[str] = None,
        SNOWFLAKE_ACCOUNT: Optional[str] = None,
        SNOWFLAKE_WAREHOUSE: Optional[str] = None,
        SNOWFLAKE_ROLE: Optional[str] = None,
        SNOWFLAKE_PASSWORD: Optional[str] = None,
        SNOWFLAKE_SOURCE_DATABASE: Optional[str] = None,
        SNOWFLAKE_DEST_DATABASE: Optional[str] = None,
    ):
        self.SNOWFLAKE_USER = SNOWFLAKE_USER or os.environ.get("SNOWFLAKE_USER")
        self.SNOWFLAKE_ACCOUNT = SNOWFLAKE_ACCOUNT or os.environ.get(
            "SNOWFLAKE_ACCOUNT"
        )
        self.SNOWFLAKE_WAREHOUSE = SNOWFLAKE_WAREHOUSE or os.environ.get(
            "SNOWFLAKE_WAREHOUSE"
        )
        self.SNOWFLAKE_ROLE = SNOWFLAKE_ROLE or os.environ.get("SNOWFLAKE_ROLE")
        self.SNOWFLAKE_PASSWORD = SNOWFLAKE_PASSWORD or os.environ.get(
            "SNOWFLAKE_PASSWORD"
        )
        self.SNOWFLAKE_SOURCE_DATABASE = SNOWFLAKE_SOURCE_DATABASE or os.environ.get(
            "SNOWFLAKE_SOURCE_DATABASE"
        )
        self.SNOWFLAKE_DEST_DATABASE = SNOWFLAKE_DEST_DATABASE or os.environ.get(
            "SNOWFLAKE_DEST_DATABASE"
        )

        try:
            _, _ = self.SNOWFLAKE_SOURCE_DATABASE.split(".")
        except:
            raise ValueError(
                "Use a fully qualified database name (e.g. CATALOG.DATABASE)."
            )
        if self.SNOWFLAKE_DEST_DATABASE:
            try:
                _, _ = self.SNOWFLAKE_DEST_DATABASE.split(".")
            except:
                raise ValueError(
                    "Use a fully qualified database name (e.g. CATALOG.DATABASE)."
                )

        required_vars = [
            "SNOWFLAKE_USER",
            "SNOWFLAKE_ACCOUNT",
            "SNOWFLAKE_WAREHOUSE",
            "SNOWFLAKE_ROLE",
            "SNOWFLAKE_SOURCE_DATABASE",
        ]
        self._check_env_vars(required_vars)
        self._check_source_dest()
        self.source_connection = self.connect_source()
        if self.SNOWFLAKE_DEST_DATABASE:
            self.dest_connection = self.connect_dest()
        else:
            self.dest_connection = None

    def _check_env_vars(self, required_vars: List[str]):
        for var in required_vars:
            if not getattr(self, var):
                raise ValueError(
                    f"Missing required variable: {var}. Set in the environment or pass through __init__()."
                )

    def _check_source_dest(self):
        if self.SNOWFLAKE_DEST_DATABASE and (
            self.SNOWFLAKE_SOURCE_DATABASE == self.SNOWFLAKE_DEST_DATABASE
            and self.SNOWFLAKE_SOURCE_SCHEMA == self.SNOWFLAKE_DEST_SCHEMA
        ):
            raise ValueError("Source and destination locations cannot be the same.")

    def _connect(self, database) -> BaseBackend:
        """
        Private method to get a database connection. End users should use connect_source() and connect_dest() to get connections to source and destination databases.
        """
        database, schema = database.split(".")
        #
        # In Ibis speak: catalog = collection of databases
        #                database = collection of tables
        #                schema = columns and column types
        # In snowflake speak: database = collection of schemas = ibis catalog
        #                schema = collection of tables = ibis database
        #
        # In the below connect method, the arguments are the SNOWFLAKE terms.
        #

        if self.SNOWFLAKE_PASSWORD:
            return ibis.snowflake.connect(
                user=self.SNOWFLAKE_USER,
                password=self.SNOWFLAKE_PASSWORD,
                account=self.SNOWFLAKE_ACCOUNT,
                warehouse=self.SNOWFLAKE_WAREHOUSE,
                role=self.SNOWFLAKE_ROLE,
                database=database,
                schema=schema,
            )
        else:
            return ibis.snowflake.connect(
                user=self.SNOWFLAKE_USER,
                authenticator="externalbrowser",
                account=self.SNOWFLAKE_ACCOUNT,
                warehouse=self.SNOWFLAKE_WAREHOUSE,
                role=self.SNOWFLAKE_ROLE,
                database=database,
                schema=schema,
            )

    def connect_dest(self) -> BaseBackend:
        """
        Establishes and returns an Ibis backend connection to the destination Snowflake database.

        Returns:
            BaseBackend: Ibis backend connection to the destination Snowflake database.
        """
        return self._connect(
            database=self.SNOWFLAKE_DEST_DATABASE,
        )

    def connect_source(self) -> BaseBackend:
        """
        Establishes and returns an Ibis backend connection to the source Snowflake database.

        Returns:
            BaseBackend: Ibis backend connection to the source Snowflake database.
        """
        return self._connect(
            database=self.SNOWFLAKE_SOURCE_DATABASE,
        )

    def get_source_table(self, name_table):
        """
        Retrieves a table from the source Snowflake database.

        Args:
            name_table (str): Name of the table to retrieve.

        Returns:
            Table: Ibis table object from the source Snowflake database.
        """
        return self.source_connection.table(
            name_table, database=self.SNOWFLAKE_SOURCE_DATABASE
        )

    def get_dest_table(self, name_table):
        """
        Retrieves a table from the destination Snowflake database.

        Args:
            name_table (str): Name of the table to retrieve.

        Returns:
            Table: Ibis table object from the destination Snowflake database.
        """
        if self.dest_connection is None:
            raise ValueError("Must specify SNOWFLAKE_DEST_DATABASE!")
        return self.dest_connection.table(
            name_table, database=self.SNOWFLAKE_DEST_DATABASE
        )

    def _get_output_table_name(self, table):
        if table.has_name:
            name_table = table.get_name().split(".")[-1]
        else:
            raise ValueError("Must specify name_table!")
        return name_table

    def create_view(self, table, name_table=None, overwrite=False):
        """
        Create a view of a table in the destination Snowflake database.

        Args:
            table (Table): Ibis table object to create a view from.
            name_table (str, optional): Name of the view to create. Defaults to None.
            overwrite (bool, optional): Whether to overwrite the view if it exists. Defaults to False.

        Returns:
            View: Ibis view object created in the destination Snowflake database.
        """
        if self.dest_connection is None:
            raise ValueError("Must specify SNOWFLAKE_DEST_DATABASE!")
        name_table = name_table or self._get_output_table_name(table)

        # Check if the destination database exists, if not, create it
        catalog, database = self.SNOWFLAKE_DEST_DATABASE.split(".")
        if not database in self.dest_connection.list_databases():
            self.dest_connection.create_database(name=database, catalog=catalog)

        return self.dest_connection.create_view(
            name=name_table.upper(),
            database=database,
            obj=table,
            overwrite=overwrite,
            schema=table.schema(),
        )

    def create_table(self, table, name_table=None, overwrite=False):
        """
        Materialize a table in the destination Snowflake database.

        Args:
            table (Table): Ibis table object to materialize.
            name_table (str, optional): Name of the table to create. Defaults to None.
            overwrite (bool, optional): Whether to overwrite the table if it exists. Defaults to False.

        Returns:
            Table: Ibis table object created in the destination Snowflake database.
        """
        if self.dest_connection is None:
            raise ValueError("Must specify SNOWFLAKE_DEST_DATABASE!")

        name_table = name_table or self._get_output_table_name(table)

        # Check if the destination database exists, if not, create it
        catalog, database = self.SNOWFLAKE_DEST_DATABASE.split(".")
        if not database in self.dest_connection.list_databases():
            self.dest_connection.create_database(name=database, catalog=catalog)

        return self.dest_connection.create_table(
            name=name_table.upper(),
            database=database,
            obj=table,
            overwrite=overwrite,
            schema=table.schema(),
        )

    def drop_table(self, name_table):
        """
        Drop a table from the destination Snowflake database.

        Args:
            name_table (str): Name of the table to drop.

        Returns:
            None
        """
        if self.dest_connection is None:
            raise ValueError("Must specify SNOWFLAKE_DEST_DATABASE!")
        return self.dest_connection.drop_table(
            name=name_table, database=self.SNOWFLAKE_DEST_DATABASE
        )

    def drop_view(self, name_table):
        """
        Drop a view from the destination Snowflake database.

        Args:
            name_table (str): Name of the view to drop.

        Returns:
            None
        """
        if self.dest_connection is None:
            raise ValueError("Must specify SNOWFLAKE_DEST_DATABASE!")
        return self.dest_connection.drop_view(
            name=name_table, database=self.SNOWFLAKE_DEST_DATABASE
        )


class DuckDBConnector:
    """
    DuckDBConnector manages connections to DuckDB using Ibis.

    Attributes:
        DUCKDB_SOURCE_DATABASE: Source DuckDB database name.
        DUCKDB_DEST_DATABASE: Destination DuckDB database name. If not specified, defaults to an in-memory database.

    Methods:
        connect_source() -> BaseBackend:
            Establishes and returns an Ibis backend connection to the source DuckDB database.

        connect_dest() -> BaseBackend:
            Establishes and returns an Ibis backend connection to the destination DuckDB database.

        get_source_table(name_table: str):
            Retrieves a table from the source DuckDB database.

        get_dest_table(name_table: str):
            Retrieves a table from the destination DuckDB database.

        create_view(table, name_table: Optional[str] = None, overwrite: bool = False):
            Create a view of a table in the destination DuckDB database.

        create_table(table, name_table: Optional[str] = None, overwrite: bool = False):
            Materialize a table in the destination DuckDB database.

        drop_table(name_table: str):
            Drop a table from the destination DuckDB database.

        drop_view(name_table: str):
            Drop a view from the destination DuckDB database.

    Example:
        ```python
        # Initialize the DuckDBConnector with a source duckdb database file with path "file.duckdb" and in-memory destination database.
        con = DuckDBConnector(DUCKDB_SOURCE_DATABASE="file.duckdb")
        ```

    Example:
        ```python
        # Initialize the DuckDBConnector with a source duckdb database file with path "source_file.duckdb" and a destination duckdb database with path "destination_file.duckdb".
        con = DuckDBConnector(
            DUCKDB_SOURCE_DATABASE="source_file.duckdb"
            DUCKDB_DEST_DATABASE="destination_file.duckdb"
        )
        ```
    """

    def __init__(
        self,
        DUCKDB_SOURCE_DATABASE: Optional[str] = None,
        DUCKDB_DEST_DATABASE: Optional[str] = None,
    ):
        """
        Initializes the DuckDBConnector with the specified path.

        Args:
            DUCKDB_SOURCE_DATABASE (str, optional): Path to the source DuckDB database.
            DUCKDB_DEST_DATABASE (str, optional): Path to the destination DuckDB database. If not specified, defaults to an in-memory database.
        """
        self.DUCKDB_SOURCE_DATABASE = DUCKDB_SOURCE_DATABASE or os.environ.get(
            "DUCKDB_SOURCE_DATABASE"
        )
        self.DUCKDB_DEST_DATABASE = (
            DUCKDB_DEST_DATABASE or os.environ.get("DUCKDB_DEST_DATABASE") or ":memory:"
        )
        required_vars = ["DUCKDB_SOURCE_DATABASE"]
        self._check_env_vars(required_vars)
        self.source_connection = self.connect_source()
        if self.DUCKDB_DEST_DATABASE:
            self.dest_connection = self.connect_dest()
        else:
            self.dest_connection = None

    def _check_env_vars(self, required_vars: List[str]):
        for var in required_vars:
            if not getattr(self, var):
                raise ValueError(
                    f"Missing required variable: {var}. Set in the environment or pass through __init__()."
                )

    def connect_source(self) -> BaseBackend:
        """
        Establishes and returns an Ibis backend connection to the source DuckDB database.

        Returns:
            BaseBackend: Ibis backend connection to the  DuckDB database.
        """
        try:
            return ibis.duckdb.connect(self.DUCKDB_SOURCE_DATABASE)
        except Exception as e:
            print(f"An error occurred: {e}")
            raise

    def connect_dest(self) -> BaseBackend:
        """
        Establishes and returns an Ibis backend connection to the destination DuckDB database.

        Returns:
            BaseBackend: Ibis backend connection to the destination DuckDB database.
        """
        if self.DUCKDB_DEST_DATABASE is None:
            raise ValueError("Must specify DUCKDB_DEST_DATABASE")
        try:
            return ibis.duckdb.connect(self.DUCKDB_DEST_DATABASE)
        except Exception as e:
            print(f"An error occurred: {e}")
            raise

    def get_source_table(self, name_table: str):
        """
        Retrieves a table from the source DuckDB database.

        Args:
            name_table (str): Name of the table to retrieve.

        Returns:
            Table: Ibis table object from the source DuckDB database.
        """
        return self.source_connection.table(name_table)

    def get_dest_table(self, name_table: str):
        """
        Retrieves a table from the destination DuckDB database.

        Args:
            name_table (str): Name of the table to retrieve.

        Returns:
            Table: Ibis table object from the destination DuckDB database.
        """
        if self.dest_connection is None:
            raise ValueError("Must specify DUCKDB_DEST_DATABASE!")
        return self.dest_connection.table(name_table)

    def create_view(self, table, name_table=None, overwrite=False):
        """
        Create a view of a table in the destination DuckDB database.

        Args:
            table (Table): Ibis table object to create a view from.
            name_table (str, optional): Name of the view to create. Defaults to None.
            overwrite (bool, optional): Whether to overwrite the view if it exists. Defaults to False.

        Returns:
            View: Ibis view object created in the destination DuckDB database.
        """
        if self.dest_connection is None:
            raise ValueError("Must specify DUCKDB_DEST_DATABASE!")

        if not name_table:
            if hasattr(table, "get_name") and table.get_name():
                name_table = table.get_name()
            else:
                raise ValueError(
                    "name_table must be provided if the table doesn't have a name."
                )

        try:
            return self.dest_connection.create_view(
                name=name_table, obj=table, overwrite=overwrite
            )
        except AttributeError as e:
            print(f"Error creating view: {e}")
            raise

    def create_table(self, table, name_table=None, overwrite=False):
        """
        Materialize a table in the destination DuckDB database.

        Args:
            table (Table): Ibis table object to materialize.
            name_table (str, optional): Name of the table to create. Defaults to None.
            overwrite (bool, optional): Whether to overwrite the table if it exists. Defaults to False.

        Returns:
            Table: Ibis table object created in the destination DuckDB database.
        """
        if self.dest_connection is None:
            raise ValueError("Must specify DUCKDB_DEST_DATABASE!")
        if not name_table:
            if hasattr(table, "get_name") and table.get_name():
                name_table = table.get_name()
            else:
                raise ValueError(
                    "name_table must be provided if the table doesn't have a name."
                )
        return self.dest_connection.create_table(
            name=name_table, obj=table, overwrite=overwrite
        )

    def drop_table(self, name_table):
        """
        Drop a table from the destination DuckDB database.

        Args:
            name_table (str): Name of the table to drop.

        Returns:
            None
        """
        if self.dest_connection is None:
            raise ValueError("Must specify DUCKDB_DEST_DATABASE!")
        self.dest_connection.drop_table(name_table)

    def drop_view(self, name_table):
        """
        Drop a view from the destination DuckDB database.

        Args:
            name_table (str): Name of the view to drop.

        Returns:
            None
        """
        if self.dest_connection is None:
            raise ValueError("Must specify DUCKDB_DEST_DATABASE!")
        self.dest_connection.drop_view(name_table)
